.leaflet-tile {
	filter: inherit;
	}

.leaflet-vml-shape {
	width: 1px;
	height: 1px;
	}
.lvml {
	behavior: url(#default#VML); 
	display: inline-block; 
	position: absolute;
	}
	
.leaflet-control {
	display: inline;
	}

.leaflet-popup-tip {
	width: 21px;
	_width: 27px;
	margin: 0 auto;
	_margin-top: -3px;
	
	filter: progid:DXImageTransform.Microsoft.Matrix(M11=0.70710678, M12=0.70710678, M21=-0.70710678, M22=0.70710678);
	-ms-filter: "progid:DXImageTransform.Microsoft.Matrix(M11=0.70710678, M12=0.70710678, M21=-0.70710678, M22=0.70710678)";
	}
.leaflet-popup-tip-container {
	margin-top: -1px;
	}
.leaflet-popup-content-wrapper, .leaflet-popup-tip {
	border: 1px solid #bbb;
	}

.leaflet-control-zoom {
	filter: progid:DXImageTransform.Microsoft.gradient(startColorStr='#3F000000',EndColorStr='#3F000000');
	}
.leaflet-control-zoom a {
	background-color: #eee;
	}
.leaflet-control-zoom a:hover {
	background-color: #fff;
	}
.leaflet-control-attribution {
	filter: progid:DXImageTransform.Microsoft.gradient(startColorstr=#B2FFFFFF,endColorstr=#B2FFFFFF);
	}