"""
Definitions for fnv_hash




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


import numpy as np
cimport numpy as np

cdef np.int64_t c_fnv_hash(unsigned unsigned char[:] octets) nogil
